#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <adwin-macros.h>
#include "adconfig.h"
#include "configfile.h"


#define CONFIG		"/proc/adwin/config"

extern int has_ip;

int add(void) {
	switch(add_dev.dev_type) {
		case DC_LINK:
			if ((signed)add_dev.dev_link_base_addr < 0) {
				fprintf(stderr, "error: link base address missing\n");
				return 1;
			}
			break;

		case DC_USB:
			if (add_dev.dev_usb_serial < 0) {
				fprintf(stderr, "error: usb serial missing\n");
				return 1;
			}
			break;
		case DC_ETHERNET:
			if (! has_ip) {
				fprintf(stderr, "error: host IP missing\n");
				return 1;
			}
			break;
		default:
			fprintf(stderr, "error: no device type specified\n");
			return 1;
	}
#ifdef KERNEL_MODULE
	int fd = open(CONFIG, O_WRONLY);
	if (fd < 0) {
		fprintf(stderr, "error: failed to open %s (%s)\n", CONFIG, strerror(errno));
		return 1;
	}
	if (ioctl(fd, ADIOKDEVADD, &add_dev) < 0) {
		fprintf(stderr, "error: ADIOKDEVADD failed (%s)\n", strerror(errno));
		close(fd);
		return 1;
	}
	close(fd);
#else // no KERNEL_MODULE
	addDeviceCF(add_dev);
#endif // KERNEL_MODULE

	return 0;
}

int del(void) {
#ifdef KERNEL_MODULE
	int fd;
	fd = open(CONFIG, O_WRONLY);
	if (fd < 0) {
		fprintf(stderr, "error: failed to open %s (%s)\n", CONFIG, strerror(errno));
		return 1;
	}
	if (ioctl(fd, ADIOKDEVDEL, del_dev) < 0) {
		fprintf(stderr, "error: ADIOKDEVDEL failed (%s)\n", strerror(errno));
		close(fd);
		return 1;
	}
	close(fd);
#else 
	PRINTDEBUG("calling deleteDeviceCF with %d \n", del_dev);
	deleteDeviceCF(del_dev);
#endif // KERNEL_MODULE
	return 0;
}


#include <net/if.h>
#include <net/if_arp.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>

static int config_device (struct in_addr broadcast_addr, int sock);	// Hilfsfunktion fr int config ()

extern struct yy_config config_dev;
extern int has_gateway, has_mask;

int config (void)
{
	struct ifconf ifc;
	struct ifreq *ifr;
	int i, result, sock, on = 1;
	char buff[1024];
	struct sockaddr_in servaddr, brdaddr;
	char *s;
	
	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock == -1)
		perror("socket");
	
	result = setsockopt(sock, SOL_SOCKET, SO_BROADCAST, &on, sizeof(on));
	if (result < 0)
		perror("setsockopt");

/*  
    // This code loops through all interfaces, and sends an open
    // broadcast to each network interface.

    // Just sending a broadcast to 255.255.255.255 does not work
    // because that broadcast is only sent to the default route (?).

    // In case of multiple configurations on a single NIC (eth0:extra
    // constructs) we are currently sending packets to each
    // configuration. It doesn't hurt, but should probably be fixed.
*/

	ifc.ifc_len = sizeof (buff);
	ifc.ifc_buf = buff;
	ifr = ifc.ifc_req;
	if (ioctl (sock, SIOCGIFCONF, &ifc) < 0) {
		fprintf (stderr, "error: Reading network interface configuration failed\n");
		return 1;
	}

	ifr = ifc.ifc_req;
	for (i = ifc.ifc_len / sizeof (struct ifreq); --i >= 0; ifr++) {
		close (sock);
		
		sock = socket(AF_INET, SOCK_DGRAM, 0);
		if (sock == -1)
			perror("socket");


		result = setsockopt(sock, SOL_SOCKET, SO_BROADCAST, &on, sizeof(on));
		if (result < 0)
			perror("setsockopt");

		
		memcpy(&servaddr, &ifr->ifr_addr, sizeof (servaddr));
		s = inet_ntoa(servaddr.sin_addr);

		if (strcmp (s, "127.0.0.1") == 0) 
			continue;

		if (bind(sock, (struct sockaddr *) &servaddr, sizeof(struct sockaddr_in)))
			perror("bind");
		brdaddr.sin_addr.s_addr = 0xffffffff;

		if (config_device (brdaddr.sin_addr, sock)) return 1;
	}
	close (sock);
	return 0;
}

static int config_device (struct in_addr broadcast_addr, int sock)
{
	struct sockaddr_in brdcast;
	int e;
	char errstr[1000];

	if (! has_ip) {
		if (! config_dev.dhcp) {
			fprintf(stderr, "error: ip missing\n");
			return 1;
		} else
			memset(config_dev.ip, '0', sizeof(config_dev.ip));
	} else if (! has_mask) {
		if (! config_dev.dhcp) {
			fprintf(stderr, "error: mask missing\n");
			return 1;
		} else
			memset(config_dev.mask, '0', sizeof(config_dev.ip));
	} else if (! has_gateway) {
		/* Frag Thomas, Stichwort: Gerteduplizierung in Scan-tool*/
		config_dev.gateway[0] = 1;
		config_dev.gateway[1] = 0;
		config_dev.gateway[2] = 0;
		config_dev.gateway[3] = 0;
	}

	/* broadcast */
	brdcast.sin_addr.s_addr = broadcast_addr.s_addr;
	brdcast.sin_family      = AF_INET;
	brdcast.sin_port        = htons (7000);
	memset (&brdcast.sin_zero, 0, sizeof (brdcast.sin_zero));

	e = sendto(sock, &config_dev, sizeof (config_dev), 0, 
		   (struct sockaddr*)&brdcast, sizeof (brdcast));
	if (e == -1)
	{
		sprintf (errstr, "Socket error %d (%s)\n", errno, strerror (errno));
		fprintf (stderr, errstr, errno, strerror (errno));
		return 1;
	}

	return 0;
}
