/***************************************************************************
                          ADEthernet.c  -  description
                             -------------------
    begin               : Wed Apr 4 2001
    copyright           : (C) 2001 by Jaeger Computergesteuerte Messtechnik GmbH
    email               : info@adwin.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <adwin-macros.h>
#include "configfile.h"

#if !defined (WIN32) && !defined (__BORLANDC__)
	#include <unistd.h>
	#include <sys/socket.h>
	#include <sys/un.h>
	#include <netinet/in.h>
	#include <arpa/inet.h>
	#include <sys/time.h>
	#include <pthread.h>

	#define CLOSE(x) close (x)
	#define SLEEP(x) usleep (x * 1000)
	#define SOCKET_TYPE int
	#define RESET_TIMEOUT(t) {(t).tv_sec  =  D->timeout / 1000; \
		                  (t).tv_usec = (D->timeout % 1000)*1000000; }

#else
	#include <winsock2.h>

	#define CLOSE(x) closesocket(x)
	#define SLEEP(x) Sleep (x)
	#define SOCKET_TYPE SOCKET
	#define RESET_TIMEOUT(t)
#endif	// !defined (WIN32) && !defined (__BORLANDC__)

// adwin-includes
#include "UDPADwin.h"
#include "ADEthernet.h"
#include "FehlerText.h"

// Parameter, die zu einem ADwin-System gehren (ADEthernet.c kennt nur 1 Device pro Thread gleichzeitig)
typedef struct
{
	SOCKET_TYPE sockfd;
	struct sockaddr_in client, server;
	unsigned int timeout;
	char passwd[10];
	int hostlink;
	long fehlercode;
	unsigned long udp_paket_count;
	long ip_version;
} ADwinDevice;

#if !defined (WIN32) && !defined (__BORLANDC__)

	static pthread_once_t DeviceOnceControl = PTHREAD_ONCE_INIT;
	static pthread_key_t DeviceKey;

	void FreeDeviceKey (void *arg)
	{
		CLOSE (((ADwinDevice*)arg)->sockfd);
		if (arg) free (arg);
	}

	void InitDeviceKey ()
	{
		pthread_key_create (&DeviceKey, FreeDeviceKey);
	}

	static inline ADwinDevice* GetCurrentDevice ()
	{
		ADwinDevice *CurrentDevice;

		pthread_once (&DeviceOnceControl, InitDeviceKey);
		CurrentDevice = pthread_getspecific (DeviceKey);
		if (!CurrentDevice)
		{
			CurrentDevice = malloc (sizeof (ADwinDevice));
			pthread_setspecific (DeviceKey, CurrentDevice);
		}

		return CurrentDevice;
	}

	#define D (GetCurrentDevice ())
#else
	ADwinDevice CurrentDevice;

	#define D (&CurrentDevice)
#endif

static long	intern_Get_Version_TcpIp();
static short Get_Data_fastshort(long, void*, short, long, long);

// Globale Variablen

//UDPH1 h1;	int ret;

//UDPR1 r1;
//UDPR2 r2;
//UDPR3 r3;

short english=1;

#define DLLVERSION 3000445
#define PLATFORM 0x80

int openConnect(char IP[16], char password[10], int link, int Port, unsigned UDPTimeout)
{

	PRINTDEBUG("openConnect Params: IP = %s, password = %s, link = %d\n", IP, password, link);
	int lauf, serversize;
#if defined (WIN32) || defined (__BORLANDC__)
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;

	wVersionRequested = MAKEWORD( 2, 2 );

	err = WSAStartup( wVersionRequested, &wsaData );
	if ( err != 0 )
	{
		return 1;
	}

	if ( LOBYTE( wsaData.wVersion ) != 2 || HIBYTE( wsaData.wVersion ) != 2 )
	{
	    WSACleanup( );
		return 1;
	}
#endif	// defined (WIN32) || defined (__BORLANDC__)

	// timeouts
	D->timeout  =  UDPTimeout;
	//h1.timeout = UDPTimeout;

	// Passwort uebernehmen
	//strcpy(h1.password , passwordd);
	strncpy(D->passwd, password, 10);

	// Link
	//h1.link = link;
	D->hostlink= link;

	// Socket kreieren
	D->sockfd = socket(AF_INET, SOCK_DGRAM, 0);
	if (D->sockfd == -1) return(1);

	// Paket-Zhler initialisieren
	D->udp_paket_count = 1;

  // Angabe zum Client
	D->client.sin_family = AF_INET;
	D->client.sin_port = 0;//htons(Port);
	D->client.sin_addr.s_addr = INADDR_ANY;

	// Angaben zum Server
	D->server.sin_family = AF_INET;
	D->server.sin_port = htons(Port);
#if !defined (WIN32) && !defined (__BORLANDC__)
	bzero(&(D->server.sin_zero), 8);
	inet_aton(IP, &D->server.sin_addr);
#else
	ZeroMemory (&(D->server.sin_zero), 8);
	D->server.sin_addr.S_un.S_addr = inet_addr (IP);
	if (D->server.sin_addr.S_un.S_addr == INADDR_NONE) return 1;
#endif
	serversize = sizeof(D->server);

	// bind
  	//if (bind(sockfd, (struct sockaddr_in *) &client, sizeof(client)) == -1) return(1);
	lauf = 100000;
	setsockopt(D->sockfd,SOL_SOCKET,SO_SNDBUF ,(const char *)&lauf,sizeof(int));
	lauf = 100000;
	setsockopt(D->sockfd,SOL_SOCKET,SO_RCVBUF,(const char *)&lauf,sizeof(int));
	D->ip_version = intern_Get_Version_TcpIp ();
	D->fehlercode = 0;
	PRINTDEBUG("end openConnect\n");
  return(0);
}

int reConnect (char *IP, char password[10], int link, int Port, unsigned UDPTimeout)
{

	PRINTDEBUG("reConnect mit IP = %s, password = %s, link = %d\n", 
		   IP, password, link);
	// Passwort uebernehmen
	//strcpy(h1.password , passwordd);
	strncpy(D->passwd, password, 10);

	// Link
	//h1.link = link;
	D->hostlink= link;

	// Paket-Zhler initialisieren
	D->udp_paket_count = 1;

#if !defined (WIN32) && !defined (__BORLANDC__)
	bzero(&(D->server.sin_zero), 8);
	inet_aton(IP, &D->server.sin_addr);
	D->server.sin_port = htons (Port);
#else
	ZeroMemory (&(D->server.sin_zero), 8);
	D->server.sin_addr.S_un.S_addr = inet_addr (IP);
	D->server.sin_port = htons (Port);
	if (D->server.sin_addr.S_un.S_addr == INADDR_NONE) return 1;
#endif

	D->ip_version = intern_Get_Version_TcpIp ();
	D->fehlercode = 0;
	D->timeout = UDPTimeout;
	PRINTDEBUG("end reConnect\n");
	return(0);
}


void closeConnect()
{
	// Socket schliessen
	CLOSE (D->sockfd);

#if defined (WIN32) || defined (__BORLANDC__)
	WSACleanup ();
#endif
}


//TODO: hier kann man doch FD_ZERO,etc weglassen, oder? ansonsten OK
int sendtoServer(int msocket, struct sockaddr_in *server, UDPH1 *h1,struct timeval *tv)
{
    int res = k_UDPtimeout;
	int count;									// neu 12.11.03

    fd_set rset;

	// neu 12.11.03
	h1->dummy1 = PLATFORM;
	h1->dummy2 = 0;
	h1->DLLVersion = DLLVERSION;

	count = sizeof(UDPH1_alt);
	if (D->ip_version <260)
	{
		if (D->ip_version >= 252)
		{
			count = sizeof(UDPH1);
		}
	}
	else
	{
		if (D->ip_version >= 264)
		{
			count = sizeof(UDPH1);
		}
	}
	// ENDE neu 12.11.03
	FD_ZERO(&rset);
	FD_SET((SOCKET_TYPE)msocket, &rset);

    if (select(msocket+1, NULL, &rset, NULL, tv) > 0)
    	res = sendto(msocket, (char *) h1, count, 0, (struct sockaddr *) server, sizeof(struct sockaddr_in));

    return res;
}


int sendtoServer3(SOCKET_TYPE msocket, struct sockaddr_in *server, UDPR3 *r3, struct timeval *tv)
{
	int res = k_UDPtimeout;

  fd_set rset;

	FD_ZERO(&rset);
	FD_SET(msocket, &rset);

  if (select(msocket+1, NULL, &rset, NULL, tv) > 0)
    res = sendto(msocket, (char *) r3, sizeof(UDPR3), 0, (struct sockaddr *) server, sizeof(struct sockaddr_in));

  return res;
}


int recvfromServer(SOCKET_TYPE msocket,UDPR1 * r1,struct timeval *tv)
{
	int res = -1;
  int count;
  fd_set rset;

	count = 0;

nochmal_recvfromServer:;

	r1->status = 0;					// Neu 10.02.04

	FD_ZERO(&rset);
	FD_SET(msocket, &rset);

	if (select(msocket+1, &rset, NULL, NULL, tv) > 0)
	{
  	res = recvfrom(msocket, (char *) r1, sizeof(UDPR1), 0, NULL, 0);
		if (res == sizeof(UDPR1))
		{
			if ((unsigned long)(r1->paket_index) != D->udp_paket_count) goto nochmal_recvfromServer;
			res = k_ok;

			if (r1->status == k_TryLater)		// gendert am 27.08.03
				res = k_TryLater;
		}
		else
		{
			if (count < k_max_retry_udp	)
			{
				count = count + 1;
				goto nochmal_recvfromServer;
			}
			res = k_WrongSize;
		}
	}
	else
		res = k_UDPtimeout;
    return(res);
}


int recvfromServer2(SOCKET_TYPE msocket, UDPR2 * r2, struct timeval *tv)
{
  int res = -1;
	int count;
	fd_set rset;
	count = 0;

nochmal_recvfromServer2:;

	r2->status = 0;					// Neu 10.02.04

	FD_ZERO(&rset);
	FD_SET(msocket, &rset);
	RESET_TIMEOUT (*tv)

	if (select(msocket+1, &rset, NULL, NULL, tv) > 0)
	{
		res = recvfrom(msocket, (char *) r2, sizeof(UDPR2), 0, NULL, 0);
		if (res == sizeof(UDPR2))
		{
			if ((unsigned long)(r2->paket_index) != D->udp_paket_count) goto nochmal_recvfromServer2;
			res = k_ok;
		}
		else
		{
			if (res == sizeof(UDPR1))						// enigefgt am 15.05.03
			{
				if (r2->status == k_TryLater)
				{
					if ((unsigned long)(r2->paket_index) == D->udp_paket_count)
					{
						return k_TryLater;				// gendert am 27.08.03
						//return k_ok;
					}
				}
			}
			if (count < k_max_retry_udp)
			{
				count = count + 1;
				goto nochmal_recvfromServer2;
			}
			res = k_WrongSize;
		}
	}
	else
	{
		res = k_UDPtimeout;
	}
    return(res);
}


int recvfromServer3(SOCKET_TYPE msocket, UDPR3 * r3, struct timeval *tv)
{
  int res = -1;
	int count;
	fd_set rset;
	count = 0;

nochmal_recvfromServer3:;

	FD_ZERO(&rset);
	FD_SET(msocket, &rset);

	if (select(msocket+1, &rset, NULL, NULL, tv) > 0)
	{
		res = recvfrom(msocket, (char *) r3, sizeof(UDPR3), 0, NULL, 0);
		if (res == sizeof(UDPR3))
		{
			if ((unsigned long)(r3->paket_index) != D->udp_paket_count) goto nochmal_recvfromServer3;
			res = k_ok;
		}
		else
		{
			if (res == sizeof(UDPR1))
			{
				if ((unsigned long)(r3->PacketNr) == D->udp_paket_count) // Vorsicht ist Tatschlich
				{											// Paket UDPR1
					return r3->paket_index;
				}
			}
			if (count < k_max_retry_udp)
			{
				count = count + 1;
				goto nochmal_recvfromServer3;
			}
			res = k_WrongSize;
		}
	}
	else
		res = k_UDPtimeout;
  return res;
}

int recvfromServer4(SOCKET_TYPE msocket,UDPR4 *r4,struct timeval *tv)
{
    int res = -1;

	int count;

	fd_set rset;
	//struct timeval tv = { MAX_READ_WAIT, 0 };

	count = 0;
nochmal_recvfromServer4:;

	r4->status = 0;						// Neu 10.02.04

	FD_ZERO(&rset);
	FD_SET(msocket, &rset);
	RESET_TIMEOUT (*tv)

	if (select(msocket+1, &rset, NULL, NULL, tv) > 0)
	{
    	res = recvfrom(msocket, (char *) r4, sizeof(UDPR4), 0, NULL, 0);
		if (res == sizeof(UDPR4))
		{
			if ((unsigned long)(r4->paket_index )!= D->udp_paket_count)
			{
				goto nochmal_recvfromServer4;
			}
			res = k_ok;
		}
		else
		{
			if (res == sizeof(UDPR1))						// enigefgt am 15.05.03
			{
				if (r4->status == k_TryLater)
				{
					if ((unsigned long)(r4->paket_index) == D->udp_paket_count)
					{
						return k_TryLater;					// gendert am 27.08.03
						//return k_ok;
					}
				}
			}
			if (count < k_max_retry_udp	)
			{
				count = count + 1;
				goto nochmal_recvfromServer4;
			}
			res = k_WrongSize;
		}
	}
	else
		res = k_UDPtimeout;

    return res;
}

int do_udp_command(int sockfd, UDPH1 *h1, UDPR1 *r1, struct sockaddr_in *server, struct timeval *tv)
{
	/* Liefert bei Fehler 1 */
	int res, retry;
	retry = 0;

	//// timeouts neu setzen
	//my_time.tv_sec  =  UDPTimeout / 1000;
	//my_time.tv_usec = (UDPTimeout % 1000)*1000000;

	udp_retry:;

	RESET_TIMEOUT (*tv)

	D->udp_paket_count = D->udp_paket_count + 1;
	h1->paket_index = D->udp_paket_count;
	if (k_UDPtimeout == sendtoServer(sockfd, server, h1, tv))
	{
		D->fehlercode = k_error_tcp_timout_read;
		return(1);
	}
	res = recvfromServer(sockfd, r1, tv);
	if (k_ok != res)
	{
		// Fehlerbehandlung
		if (res > 0)
		{
			D->fehlercode = res;
			return(1);
		}
		if (r1->status == k_TryLater)
		{
			SLEEP (10);
			goto udp_retry;
		}
		D->fehlercode = k_error_tcp_error;
		if (res == k_UDPtimeout)
		{
			D->fehlercode = k_error_tcp_timout_read;
		}
		if (retry < k_max_retry_udp)
		{
			retry = retry + 1;
			goto udp_retry;
		}
		return(1);
	}
	if (r1->status == k_WrongPassword)
	{
		D->fehlercode = k_error_tcp_password;
		return(1);
	}
	if (r1->status != k_ok)
	{
		D->fehlercode = r1->status;
		return(1);
	}
	return(0);
}


int do_udp_command2(int sockfd, UDPH1 *h1, UDPR2 *r2, struct sockaddr_in *server, struct timeval *tv)
{
	int res, retry;
	retry = 0;

	//// timeouts neu setzen
	//my_time.tv_sec  =  UDPTimeout / 1000;
	//my_time.tv_usec = (UDPTimeout % 1000)*1000000;


udp_retry:;

	D->udp_paket_count = D->udp_paket_count + 1;
	h1->paket_index = D->udp_paket_count;

	if (k_UDPtimeout == sendtoServer(sockfd, server, h1, tv))
	{
		D->fehlercode = k_error_tcp_timout_read;
		return(1);
	}
	res = recvfromServer2(sockfd, r2, tv);
	if (k_ok != res)
	{
		// Fehlerbehandlung
		if (res > 0)
		{
			D->fehlercode = res;
			return(1);
		}
		if (r2->status == k_TryLater)
		{
			SLEEP (10);
			goto udp_retry;
		}
		D->fehlercode = k_error_tcp_error;
		if (res == k_UDPtimeout)
		{
			D->fehlercode = k_error_tcp_timout_read;
		}
		if (retry  < k_max_retry_udp)
		{
			retry = retry + 1;
			goto udp_retry;
		}
		return(1);
	}
	if (r2->status == k_WrongPassword)
	{
		D->fehlercode = k_error_tcp_password;
		return(1);
	}
	if (r2->status != k_ok)
	{
		D->fehlercode = r2->status;
		return(1);
	}
	D->fehlercode = 0;
	return (0);
}

int do_udp_command_nx4(SOCKET_TYPE sockfd, UDPH1 *h1, UDPR4 *r4, struct sockaddr_in *server, struct timeval *tv,
					   int blockcount, char *buffer, short *datatyp)
{
	int res,retry,lauf;

	retry = 0;
udp_retry:;
	D->udp_paket_count = D->udp_paket_count + 1;
	h1->paket_index = D->udp_paket_count;

	if (k_UDPtimeout == sendtoServer(sockfd, server, h1, tv))
	{
		D->fehlercode = k_error_tcp_timout_read;
		return(1);
	}

	RESET_TIMEOUT (*tv)

	for (lauf = 0; lauf < blockcount; lauf++)
	{
		res = recvfromServer4(sockfd,r4,tv);
		if (k_ok != res)
		{
			// Fehlerbehandlung

			if (res > 0)
			{
				D->fehlercode = res;
				return(lauf);
			}

			if (r4->status == k_TryLater)
			{
				SLEEP (10);
				goto udp_retry;
			}

			D->fehlercode = k_error_tcp_error;

			if (res == k_UDPtimeout)
			{
				D->fehlercode = k_error_tcp_timout_read;
			}
			if (retry < k_max_retry_udp)
			{
				retry = retry + 1;
				goto udp_retry;
			}
			return(lauf);
		}
		if (r4->status == k_WrongPassword)
		{
			D->fehlercode = k_error_tcp_password;
			return(lauf);
		}

		if (r4->status != k_ok)
		{
			D->fehlercode = r4->status;
			return(lauf);
		}
		if (lauf == 0)
			datatyp[0] = (short)(r4->datatyp);
		if (r4->PacketNr != lauf)
		{
			D->fehlercode = k_error_tcp_timout_read;
			return lauf;
		}
		memcpy(&buffer[lauf * 350 * 4],r4->value,350 * 4);
	}

#ifndef _MYUDPTEST
	D->fehlercode = 0;
#endif //_MYUDPTEST
	return (blockcount);
}

long NAME (Get_Par) (long Index)
{

	UDPH1 h1;			//Neu
	UDPR1 r1;			//Neu
	struct timeval my_time; //Neu

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	h1.timeout = D->timeout;		//Neu
	h1.link= D->hostlink;				//Neu
	strcpy(h1.password, D->passwd);	//Neu

	h1.befehl   = k_3plus1;
	h1.value[0] = k_GetPar;
	h1.value[1] = 1000 + Index;
	h1.value[2] = 0;

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) return(1); //false

    //*value = r1.value[0];
	//return(0);
	return r1.value[0];
}


void NAME (Set_Par) (long Index, long Value)
{
	UDPH1 h1;			//Neu
	UDPR1 r1;			//Neu
	struct timeval my_time; //Neu

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	h1.timeout = D->timeout;		//Neu
	h1.link= D->hostlink;				//Neu
	strcpy(h1.password, D->passwd);	//Neu

	h1.befehl   = k_3plus1;
	h1.value[0] = k_SetPar;
	h1.value[1] = 1000 + Index;
	h1.value[2] = Value;

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) return;	//(1)//false bei M.G.
	return;
}


void NAME (Set_FPar) (long Index, float Value)
{
	UDPH1 h1;			//Neu
	UDPR1 r1;			//Neu
	struct timeval my_time; //Neu
	long *lptr;

	lptr = (void*)&Value;

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	h1.timeout = D->timeout;		//Neu
	h1.link= D->hostlink;				//Neu
	strcpy(h1.password, D->passwd);	//Neu

	h1.befehl   = k_3plus1;
	h1.value[0] = k_SetPar;
	h1.value[1] = 1100 + Index;
	h1.value[2] = *lptr;

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) return; //(1);
	return;//(0);
}


float NAME (Get_FPar) (long Index)
{
	UDPH1 h1;			//Neu
	UDPR1 r1;			//Neu
	struct timeval my_time; //Neu

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	h1.timeout = D->timeout;		//Neu
	h1.link= D->hostlink;				//Neu
	strcpy(h1.password, D->passwd);	//Neu

	h1.befehl   = k_3plus1;
	h1.value[0] = k_GetPar;
	h1.value[1] = 1100 + Index;
	h1.value[2] = 0;

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time))
	{
		//// Fehler
		//*value = 0;
		return(1); //false
	}
	else
	{
		//*value= *((float*)(&r1.value[0]));
		return *((float*)(&r1.value[0]));//(0);
  }
}


short NAME (GetDataType) (short datanr)
{
	UDPH1 h1;			//Neu
	UDPR1 r1;			//Neu
	struct timeval my_time; //Neu

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	h1.timeout = D->timeout;		//Neu
	h1.link= D->hostlink;				//Neu
	strcpy(h1.password, D->passwd);	//Neu

	// 0 Fehler oder nicht deklariert, 3 long, 5 float
	h1.befehl   = k_GetDataType;
	h1.value[0] = k_GetData;
	h1.value[1] = datanr;
	h1.value[2] = 1;
	h1.value[3] = 0;

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) return(0);
	return ((short)r1.value[0]);
}


void NAME (Get_Par_All) (long *pData)
{
	UDPH1 h1;			//Neu
	UDPR2 r2;			//Neu
	struct timeval my_time; //Neu

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	h1.timeout = D->timeout;		//Neu
	h1.link= D->hostlink;				//Neu
	strcpy(h1.password, D->passwd);	//Neu

	h1.befehl   = k_GetParAll;
	h1.value[0] = 1;//np;
	h1.value[1] = 80;//anzahl;
	if (do_udp_command2(D->sockfd, &h1, &r2, &(D->server), &my_time)) return;//1;

	memcpy(pData, r2.value, 80 * sizeof(long));
	return;//(0);
}

void NAME (Get_Par_Block) (long *pData, long StartIndex, long Count)
{
	UDPH1 h1;			//Neu
	UDPR2 r2;			//Neu
	struct timeval my_time; //Neu

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	h1.timeout = D->timeout;		//Neu
	h1.link= D->hostlink;				//Neu
	strcpy(h1.password, D->passwd);	//Neu

	h1.befehl   = k_GetParAll;
	h1.value[0] = StartIndex;
	h1.value[1] = Count;

	if (do_udp_command2(D->sockfd, &h1, &r2, &(D->server), &my_time)) return;//1;
	memcpy(pData, r2.value, Count * sizeof(long));

	return;//(0);
}


void NAME (Get_FPar_All) (float *pData)
{

	UDPH1 h1;			//Neu
	UDPR2 r2;			//Neu
	struct timeval my_time; //Neu

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	h1.timeout = D->timeout;		//Neu
	h1.link= D->hostlink;				//Neu
	strcpy(h1.password, D->passwd);	//Neu

	h1.befehl   = k_GetParAll;
	h1.value[0] = 100 + 1;
	h1.value[1] = 80;

	if (do_udp_command2(D->sockfd, &h1, &r2, &(D->server), &my_time)) return;// 1;

	memcpy(pData, r2.value, 80 * sizeof(long));
	return;//(0);
}

void NAME (Get_FPar_Block) (float *pData, long StarIndex, long Count)
{

	UDPH1 h1;			//Neu
	UDPR2 r2;			//Neu
	struct timeval my_time; //Neu

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	h1.timeout = D->timeout;		//Neu
	h1.link= D->hostlink;				//Neu
	strcpy(h1.password, D->passwd);	//Neu

	h1.befehl   = k_GetParAll;
	h1.value[0] = 100 + StarIndex;
	h1.value[1] = Count;

	if (do_udp_command2(D->sockfd, &h1, &r2, &(D->server), &my_time)) return;// 1;

	memcpy(pData, r2.value, Count * sizeof(long));
	return;//(0);
}

/***************Fifos**********************************************************************************/
long NAME (Fifo_Full) (long FifoNo)
{
	UDPH1 h1;			//Neu
	UDPR1 r1;			//Neu
	struct timeval my_time; //Neu
	long lh;


	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	h1.timeout = D->timeout;		//Neu
	h1.link= D->hostlink;				//Neu
	strcpy(h1.password, D->passwd);	//Neu

	h1.befehl   = k_3plus1;
	h1.value[0] = k_GetFifoCount;
	h1.value[1] = FifoNo;
	h1.value[2] = 0;

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) return(255);

	lh = r1.value[0];
	return(lh);
}


long NAME (Fifo_Empty) (long FifoNo)
{
	UDPH1 h1;			//Neu
	UDPR1 r1;			//Neu
	struct timeval my_time; //Neu
	long lh;

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	h1.timeout = D->timeout;		//Neu
	h1.link= D->hostlink;				//Neu
	strcpy(h1.password, D->passwd);	//Neu

	h1.befehl   = k_3plus1;
	h1.value[0] = k_GetFifoEmpty;
	h1.value[1] = FifoNo;
	h1.value[2] = 0;

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) return(255);
	lh = r1.value[0];
	return(lh);
}

void NAME (Fifo_Clear) (long FifoNo)
{
	UDPH1 h1;			//Neu
	UDPR1 r1;			//Neu
	struct timeval my_time;  //Neu

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	h1.timeout = D->timeout;		//Neu
	h1.link= D->hostlink;				//Neu
	strcpy(h1.password, D->passwd);	//Neu

 	h1.befehl   = k_3plus1;
 	h1.value[0] = k_ClearFifo;
 	h1.value[1] = FifoNo;
 	h1.value[2] = 0;
 	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) return;//(1); //255 bei M.G.
 	return; // (0);
}


short Get_Fifo_UDP (char *Data, short FifoNo, long Count, short dsp_typ, long *pkt_index, short flag_retry, struct sockaddr_in *server)
{
	UDPH1 h1;					//Neu
	UDPR3 r3;					//Neu
	struct timeval my_time; //Neu
	int retry;
	int size, pointer, zaehler;

Get_Fifo_UDP_retry:														// Neu 10.02.04

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	strcpy(h1.password, D->passwd); //neu
	h1.link = D->hostlink;			 //neu
	h1.timeout = D->timeout;     //neu

	D->fehlercode = 0;

	if (flag_retry)
		h1.befehl   = k_GetFifoRetry;
	else
		h1.befehl   = k_GetFifo;


	h1.value[0] = dsp_typ;
	h1.value[1] = FifoNo | (5 << 16);		// ?
	if (flag_retry)
		h1.value[2] = *pkt_index;

	h1.value[3] = Count;

	size = Count * sizeof(long);

	retry = 0;

//udp_retry:;	 //bei M.G. enthalten, wird aber nicht aufgerufen

	D->udp_paket_count = D->udp_paket_count + 1;
	h1.paket_index = D->udp_paket_count;

	if (!flag_retry)
		*pkt_index = D->udp_paket_count;

	// senden Standardpaket
	if (k_UDPtimeout == sendtoServer(D->sockfd, server, &h1, &my_time))
	{
		D->fehlercode = k_error_tcp_timout_read;
		return(-1);
	}

	// Nun max. 9800 Byte entspricht 7 Pakete
	pointer = 0;
	zaehler = 0;
	while(size > 0)
	{
		int res;

		res = recvfromServer3(D->sockfd, &r3, &my_time);
		if (res == k_ok)
		{
			// Koennte noch Fehlercode sein

			if (r3.PacketNr == zaehler)
			{
				if (size >= sizeof(r3.value))
					memcpy(&Data[pointer],r3.value,sizeof(r3.value));
				else
					memcpy(&Data[pointer],r3.value,size);
				size = size - sizeof(r3.value);
				pointer = pointer + sizeof(r3.value);
				zaehler = zaehler + 1;
			}
		}
		else
		{
			if (res == k_TryLater)							// Neu 10.02.04
			{												// Neu 10.02.04
				SLEEP (10);									// Neu 10.02.04
				goto Get_Fifo_UDP_retry;					// Neu 10.02.04
			}												// Neu 10.02.04
			if ((res == k_RetryUnknown) && flag_retry)
				return 200;
			if (res == k_UDPtimeout)
			{
				return 255;
			}
			if (res == k_FifoLessData)
				D->fehlercode = k_Fifo_to_less_data;
			return 150;
			}
		}
	return 0;
}

#define GET_FIFO_BLOCK_SIZE 6

void NAME (Get_Fifo) (long FifoNo,short typ, void *Data,long Count)
//short Get_Fifo(void *tarr, short typ, short nadc, long il)
{
	short dsp_typ;
	char *daten;
	long pkt_index;
	long pkt_pointer,msize;
	long l1 ;
	short res;
	long  *plong;
	float *pfloat;

	if (Count < 1) return;// 0;

	res = 0;
	D->fehlercode = 0;

	dsp_typ = (short)NAME (GetDataType) ((short)FifoNo);

	if (dsp_typ == 0)          // Fehler
	{
		if (D->fehlercode == 0) D->fehlercode = k_Fifo_is_to_small;
		return;//;(255);
	}

	if ((dsp_typ == 3) || (dsp_typ == 4))		// 3 long
		dsp_typ = 2;

	if (dsp_typ != typ)
		daten = malloc(sizeof(long) * Count);
	else
		daten = Data;

	msize = Count;

	pkt_pointer = 0;

	while (msize > 0)
	{
		int counter;

		counter = 0;

nochmal_von_vorne:;

		if (msize >= GET_FIFO_BLOCK_SIZE * 350)
			res = (short)Get_Fifo_UDP (&daten[pkt_pointer], (short)FifoNo, GET_FIFO_BLOCK_SIZE * 350, dsp_typ, &pkt_index, 0, &(D->server));
		else
			res = (short)Get_Fifo_UDP (&daten[pkt_pointer], (short)FifoNo, msize, dsp_typ, &pkt_index ,0, &(D->server));

		while ((res != 0) && (counter < 5))
		{
			switch (res)
			{
				case 255:
				{
					if (msize >= GET_FIFO_BLOCK_SIZE * 350)  //9800 / sizeof(long)
						res = (short)Get_Fifo_UDP (&daten[pkt_pointer], (short)FifoNo, GET_FIFO_BLOCK_SIZE * 350, dsp_typ,  &pkt_index, 1, &(D->server));
					else
						res = (short)Get_Fifo_UDP (&daten[pkt_pointer], (short)FifoNo, msize, dsp_typ, &pkt_index ,1, &(D->server));
				}
				break;

				case 200:
				{
					/*counter = counter + 1;						// 06.05
					if (counter < 5)*/ goto nochmal_von_vorne;
				}
				break;

				case 150:
					return;//(255);

				case 100:
					D->fehlercode = k_Fifo_to_less_data;
					return;//(255);

				case -1:
					D->fehlercode = k_error_tcp_timout_read;
					return;// 255;
			}
			counter = counter + 1;

			if (counter >= 5)
			{
				D->fehlercode = k_error_tcp_timout_read;
				return;//(255);
			}
		}
		msize = msize - GET_FIFO_BLOCK_SIZE * 350;
		pkt_pointer = pkt_pointer + GET_FIFO_BLOCK_SIZE * 350 * sizeof(long);
	}

	D->fehlercode = 0;

	if (dsp_typ != typ)
	{
		switch (typ)
		{
			//TODO://case 1: 6,20

			case 2:	// 32bit int
				plong = Data;
				if (dsp_typ == 5)
				{
					pfloat =  (float *)  daten;
					for (l1=0; l1<Count  ; l1++)
						plong[l1] = (long)(pfloat [l1]);
   				}
			break;
			case 5:		//32bit float
				pfloat = Data;
				if (dsp_typ == 2)
				{
					plong =  (long *)  daten;
					for (l1=0; l1<Count  ; l1++)
					pfloat[l1] = (float)(plong[l1]);
				}
			break;
		}
		free(daten);
	}
	return;//(0);
}


short Set_Fifo_UDP (char *Data, short FifoNo, long Count ,short dsp_typ, long *pkt_index ,short flag_retry, struct sockaddr_in *server)
{
	int size, pointer, zaehler;
	UDPH1 h1;
	UDPR1 r1;
	UDPR3 r3;
	struct timeval my_time; //Neu

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	strcpy(h1.password, D->passwd); //neu
	h1.link = D->hostlink;			 //neu
	h1.timeout = D->timeout;     //neu

	D->fehlercode = 0;

	if (flag_retry)
		h1.befehl   = k_SetFifoRetry;
	else
		h1.befehl   = k_SetFifo;

	h1.value[0] = dsp_typ;
	h1.value[1] = FifoNo | (5 << 16);
	if (flag_retry)
		h1.value[2] = *pkt_index;
	h1.value[3] = Count;

	if (do_udp_command(D->sockfd, &h1, &r1, server, &my_time))
	{
		if ((flag_retry) && (r1.status == k_SchonBearbeitet))
		{
			D->fehlercode = k_ok;
			return 0;
		}
		else
			return 255;
	}

	if (!flag_retry)
		*pkt_index = D->udp_paket_count;

	// Nun max. 9800 Byte entspricht 7 Pakete
	pointer = 0;
	zaehler = 0;
	size = Count * sizeof(long);

	while(size > 0)
	{
		r3.paket_index = D->udp_paket_count;
		r3.PacketNr = zaehler;
		if (sizeof(r3.value) <=  size)
			memcpy(r3.value,&Data[pointer],sizeof(r3.value));
		else
			memcpy(r3.value,&Data[pointer],size);

		sendtoServer3(D->sockfd, server, &r3, &my_time);

		size = size - sizeof(r3.value);
		pointer = pointer + sizeof(r3.value);
		zaehler = zaehler + 1;
	}

	// Es fehlt noch letzte Handschak;

	if (recvfromServer(D->sockfd, &r1, &my_time) == k_ok)
	{
		if (r1.status == k_ok) return 0;
		if (r1.status == k_SchonBearbeitet) return 0;
		if (r1.status == k_timeout)
		{
			return 100;
		}
		if (r1.status == k_FalschePaketreihenfolge) return 100;
		return 200;
	}
	else
		return 100;
}


void NAME (Set_Fifo) (long FifoNo,short typ, void *Data,  long Count)
{
	long l1 ;
	short res;
	long  *plong;
	float *pfloat;
	int retry;
	short dsp_typ;
	int msize,pkt_pointer;
	long pkt_index;
	char *daten;
	int counter;

	if (Count < 1) return;//(0);

	//wird hier nicht gebraucht
	//UDPH1 h1;
	//UDPR1 r1;
	//struct timeval my_time; //Neu

	//my_time.tv_sec  =  UDPTimeout / 1000;		//Neu
	//my_time.tv_usec = (UDPTimeout % 1000)*1000000;	//Neu

	//strcpy(h1.password, passwd); //neu
	//h1.link = hostlink;			 //neu
	//h1.timeout = UDPTimeout;     //neu

	D->fehlercode = 0;

	dsp_typ = (short)NAME (GetDataType) ((short)FifoNo);

	if (dsp_typ == 0)
	{
		if (D->fehlercode == 0) D->fehlercode = k_Fifo_is_to_small;
		return;//(255);
	}

	if ((dsp_typ == 3) || (dsp_typ == 4))
		dsp_typ = 2;

	if (dsp_typ != typ)
	{
		daten = malloc(sizeof(long) * Count);
		switch (typ)
		{
			//TODO: case 1,6 und 20 ??
			case 2:			// 32bit int
				if (dsp_typ == 5)
				{
					plong = (long *) Data;
					pfloat =  (float *)  daten;
					for (l1=0; l1<Count  ; l1++)
					{
						pfloat [l1] = (float)(plong[l1]);
   				}
				}
				break;
			case 5:			// 32bit float
				if (dsp_typ == 2)
				{
					pfloat =  (float *)  daten;
					plong =  (long *)  daten;
					for (l1=0; l1<Count  ; l1++)
						plong[l1] = (long)(pfloat[l1]);
				}
				break;
		}
	}
	else
		daten = (char *)Data;

	retry = 0;
	msize = Count;
	pkt_pointer = 0;
	while (msize > 0)
	{
		counter = 0;

		if (msize >= 2450)  //9800 / sizeof(long)
			res = (short)Set_Fifo_UDP (&daten[pkt_pointer], (short)FifoNo, 2450, dsp_typ, &pkt_index, 0, &(D->server));
		else
			res = (short)Set_Fifo_UDP (&daten[pkt_pointer], (short)FifoNo, msize, dsp_typ, &pkt_index, 0, &(D->server));

		while ((res != 0) && (counter < 5))
		{
			switch (res)
			{
				case 100:
					{
						if (msize >= 2450)  //9800 / sizeof(long)
							res = (short)Set_Fifo_UDP (&daten[pkt_pointer], (short)FifoNo, 2450, dsp_typ, &pkt_index, 1, &(D->server));
						else
							res = (short)Set_Fifo_UDP (&daten[pkt_pointer], (short)FifoNo, msize, dsp_typ, &pkt_index, 1, &(D->server));
					}
					break;

				case 200:
					D->fehlercode = k_Fifo_is_to_small;
					return;//(255);

				case 255:
					return;// 255;
			}
			counter = counter + 1;

			if (counter >= 5)
			{
				D->fehlercode = k_error_tcp_timout_read;
				return;//(255);
			}
		}
		msize = msize - 2450;
		pkt_pointer = pkt_pointer + 2450 * sizeof(long);
	}

	D->fehlercode = 0;

	return;//(0);
}
/****Datas***********************************************************************************************/

// 1 Fehler, 0 ok

short getarray (char *arr, long msize, long typsize,SOCKET_TYPE msocket, struct sockaddr *m_serv_addr, int max_paket_count, struct timeval *tv)
{
	long size,last_size,retry,count;
	long arraypointer;
	long anzahl_empfangen;
	long last_arraypointer;
	long packet_nr;
	int fehlerblock;
	UDPR3 message_out3;
	UDPR1 message_out;
	int value1,res,retry2/*,retry3*/;
	long pNr;
	int paket_erreicht[50];
	int dummy_zaehler;
	int first_empfang;

	int counti;


	fd_set  my_fd_setl;

	counti = 0;
	first_empfang = 0;

	dummy_zaehler = 0;
	anzahl_empfangen = 0;
	fehlerblock = 0;
	arraypointer = 0;
	packet_nr = 1;
	pNr = 1;
	retry = 0;
	retry2 = 0;
	size = msize * typsize; // Gesamtgroesse Datei

	message_out.value[0] = max_paket_count;

	while(size > 0)
	{
		last_arraypointer = arraypointer;
		last_size = size;

		for (value1 = 0; value1 < max_paket_count; value1++)
			paket_erreicht[value1] = 0;


		for (value1 = 0; value1 < max_paket_count; value1++)
		{
			if (size > 0)
			{
nochmal_empfangen2:;

				FD_ZERO(&my_fd_setl);
				FD_SET(msocket,&my_fd_setl);
				RESET_TIMEOUT (*tv)

				if (1 == select(msocket+1, &my_fd_setl, 0, 0 ,tv))
				{
					struct sockaddr dummy;
					int len_dummy;

					len_dummy = sizeof(dummy);
					count = recvfrom(msocket, (char *)&message_out3, sizeof(message_out3), 0, &dummy, &len_dummy);
					if (0 < count)
					{
						{
							if (count == sizeof(message_out3))
							{
								if(message_out3.paket_index != pNr)
								{
									retry = retry + 1;
									if (retry > k_max_retry_get)
									{
										D->fehlercode = k_error_tcp_error;
										return(1);
									}
									goto nochmal_empfangen2;
								}
								else
								{
									first_empfang = 1;
									anzahl_empfangen = anzahl_empfangen + 1;
									//if (size > sizeof(message_out3.value))
										memcpy(&arr[arraypointer + message_out3.PacketNr * sizeof(message_out3.value)] , message_out3.value,sizeof(message_out3.value));
									//else
									//	memcpy(&arr[arraypointer + message_out3.PacketNr * sizeof(message_out3.value)] , message_out3.value,size);
									paket_erreicht[message_out3.PacketNr] = 1;
									size = size - sizeof(message_out3.value);
								}
							}
							else
							{
								retry = retry + 1;
								if (retry > k_max_retry_get)
								{
									D->fehlercode = k_error_tcp_error;
									return(1);
								}
								goto nochmal_empfangen2;
							}
						}
					}
					else
					{
						// Fehlerhandling falscher empfang
						retry = retry + 1;
						if (retry > k_max_retry_get)
						{
							D->fehlercode = k_error_tcp_error;
							return(1);
	}
						goto nochmal_empfangen2;
					}
				}
				else
				{
					// Fehlerhandling timeout
					retry = retry + 1;
					if (retry > k_max_retry_get)
					{
						D->fehlercode = k_error_tcp_timout_read;
						return(1);
					}
					if (anzahl_empfangen == 0) goto nochmal_senden;
					else value1 = max_paket_count ;
				}
			}
		}

		// Falls alles richtig
		fehlerblock = -1;
		for (value1 = 0; value1 < max_paket_count ; value1++)
		{
			{
				int helpi;
				helpi = last_size - value1 * sizeof(message_out3.value);
				if (helpi <= 0) helpi = last_size + 1;

				if (last_size - helpi >= 0)
				{
					if (paket_erreicht[value1] != 1)
					{
						fehlerblock = value1;
						value1 = max_paket_count;
					}
				}
			}
		}
		if (fehlerblock == -1)
		{
			message_out.status = k_ok;
			message_out.value[0] = max_paket_count;
			arraypointer = arraypointer + (last_size - size);
		}
		else
		{
			message_out.status = k_PaketFehler;
			message_out.value[0] = fehlerblock;
			size = last_size - fehlerblock * sizeof(message_out3.value);
			arraypointer = arraypointer + fehlerblock * sizeof(message_out3.value);
			if (fehlerblock == 0) retry = retry + 1;
			if (retry > k_max_retry_get)
			{
				D->fehlercode = k_error_tcp_error;
				return(1);
			}
		}
		retry = 0;
		message_out.paket_index= pNr++;
		counti = counti + 1;
nochmal_senden:;

		if (first_empfang == 0)
		{
			message_out.status = k_PaketFehler;
			message_out.value[0] = 0;
			message_out.paket_index = 0;
			retry = retry + 1;
			if (retry > k_max_retry_get)
			{
				D->fehlercode = k_error_tcp_error;
				return(1);
			}
		}

/*#ifdef _MYUDPTEST
		if (	dummy_zaehler == 7)
			dummy_zaehler  = 0;
		else
		{
#endif //_MYUDPTEST*/
			res = sendto(msocket, (char *) &message_out, sizeof(UDPR1), 0,m_serv_addr,sizeof(struct sockaddr));
			dummy_zaehler = dummy_zaehler + 1;
/*#ifdef _MYUDPTEST
		}
#endif //_MYUDPTEST*/
		anzahl_empfangen = 0;
	}

	D->fehlercode = 0;

	return(0);
}

// TRUE Fehler FALSE ok

short setarray (char *arr, long msize, long typsize,SOCKET_TYPE msocket, struct sockaddr *server, int max_paket_count, struct timeval *tv)
{

	long size,last_size,retry,zaehler,zaehler1,zaehler3;
	long arraypointer;
	long last_arraypointer;
	long packet_nr;
	int fehlerblock;
	UDPR3 message_out3;
	UDPR1 message_out;

	int value1,res,retry2,retry3,retry4;

	fd_set  my_fd_setl;

	zaehler = 0;
	zaehler1 = 1;
	zaehler3 = 0;

	fehlerblock = 0;
	arraypointer = 0;
	packet_nr = 1;
	retry = 0;
	retry2 = 0;
	size = msize * typsize; // Gesamtgroesse Datei

	while(size > 0)
	{

		last_arraypointer = arraypointer;
		last_size = size;
nochmal_senden:;
		for (value1 = 0; value1 < max_paket_count; value1++)
		{
			if (size > 0)
			{

				if (size > sizeof(message_out3.value))
					memcpy(message_out3.value,&arr[arraypointer + value1 * sizeof(message_out3.value)],sizeof(message_out3.value));
				else
					memcpy(message_out3.value,&arr[arraypointer + value1 * sizeof(message_out3.value)],size);
				message_out3.PacketNr = value1;	// von 0 bis 4
				message_out3.paket_index = zaehler1;

//#ifdef _MYUDPTEST
//				zaehler3 = zaehler3 + 1;
//				if (	zaehler3 < 17)
//				{
//#endif //_MYUDPTEST
					res = sendto(msocket, (char *) &message_out3, sizeof(message_out3), 0, server, sizeof(struct sockaddr));
//#ifdef _MYUDPTEST
//				}
//				else
//					zaehler3 = 0;
//#endif //_MYUDPTEST
//				if (res == sizeof(message_out3))
//				{
					size = size - sizeof(message_out3.value);
			}
		}

		retry3 = 0;
		retry4 = 0;
		retry = 0;
		retry2 = 0;

nochmal_empfangen:;

		FD_ZERO(&my_fd_setl);
		FD_SET(msocket,&my_fd_setl);
		RESET_TIMEOUT (*tv)

		if (1 ==select(D->sockfd+1,&my_fd_setl,0,0 ,tv))
		{
			struct sockaddr dummy;
			int len_dummy;

			len_dummy = sizeof(dummy);
			res = recvfrom(msocket, (char *)&message_out, sizeof(message_out), 0, &dummy, &len_dummy);
			if (res == sizeof(message_out))
			{

				if ((size <= 0) && (message_out.status == k_ok))
				{
					return(0); //False
				}

				if (message_out.paket_index != zaehler1)
				{
					if (message_out.paket_index == zaehler1-1)
					{
						retry4 = retry4 + 1;

						if (retry4 < 3)
							goto nochmal_empfangen;
						else
						{
							// Nochmal senden
							size = last_size;
							goto nochmal_senden;
						}
					}
					else
					{

						// Falls 1 Block und Falls paket_index = udp_paket_count

						retry4 = 0;

						if ((zaehler1 == 1) && ((unsigned long)(message_out.paket_index) == D->udp_paket_count))
						{
							message_out.status = k_PaketFehler;
							message_out.value[0] = 0;
							zaehler1 = 0;						// Neu 27.04.04  Damit es mehr als einmal Wiederholt wird !						}
						}
						else
							goto nochmal_empfangen;
					}
				}
				zaehler1 = zaehler1 + 1;
				if (message_out.status != k_ok)

				{
					if (message_out.status == k_PaketFehler)
					{
						arraypointer = last_arraypointer + message_out.value[0] * sizeof(message_out3.value);
						size = last_size - message_out.value[0] * sizeof(message_out3.value);
					}
					else
					{
						D->fehlercode = message_out.status;
						return(1);
					}
					retry = retry + 1;
					if (retry > k_max_retry)
					{
						D->fehlercode = k_error_tcp_error;
							return(1);
					}
				}
				else
					arraypointer = last_arraypointer + max_paket_count * sizeof(message_out3.value);

			}
			else
			{

				if ((size <= 0) && (retry2 == 2))
				{
					D->fehlercode = 0;
					return(0);
				}
				retry3 = retry3 + 1;
				if (retry3 < k_max_retry)  goto nochmal_empfangen;
				D->fehlercode = k_error_tcp_error;
				return(1);
			}
		}
		else
		{
			// irgentwas schiefgegangen timeout erreicht
			retry3 = retry3 + 1;
			if (retry3 < k_max_retry)  goto nochmal_empfangen;
			D->fehlercode = k_error_tcp_timout_read;
			return(1);
		}

	}
	D->fehlercode = 0;
	return(0);
}
/*
	Data 	Pointer auf Array

	typ	 	Datentyp
		1 	16 Bit GANZZAHL
		2	32 BIT GANZZAHL
		5	32 BIT FLOAT
		6	double
	   20	VARIANT

	DataNo 	Datanummer
	StartIndex	erster Wert
	Count		Anzahl
	nadr	Linkaddresse
*/

void NAME (Get_Data) (long DataNo,void *Data, short typ,  long StartIndex, long Count)
{
	long l1 ;

	long  *plong;
	float *pfloat;

	int pointer,start,msize;
	short dsp_typ;
	char *daten;

	UDPH1 h1;			//Neu
	UDPR1 r1;			//Neu
	struct timeval my_time;  //Neu

	if (Count < 1) return;//(1);//0 bei M.G.

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	D->fehlercode = 0;

	//// falls klein und Version > als
	//// 28.01.02 nderungen wegen Jrgen Version von 300 auf 250 zurckgeasetzt,
	if (( Count <= 5 * 350) && (D->ip_version >= 250))
	{
		D->fehlercode = Get_Data_fastshort(DataNo, Data, typ, StartIndex, Count);
		return;
	}

	dsp_typ = (short)NAME (GetDataType) ((short)DataNo);

	if (dsp_typ == 0)
	{
		if (D->fehlercode == 0) D->fehlercode = k_Data_is_to_small;
		if (D->fehlercode == k_DataZuKlein) D->fehlercode = k_Data_is_to_small;
		return;//(1); //255 bei M.G.
	}

 	if ((dsp_typ == 3) || (dsp_typ == 4))
 		dsp_typ = 2;

 	//if (dsp_typ != typ)										//Neu
 	{
		unsigned long helpi;									// Neu
		long blksize;											// Neu

		blksize = 5;					// Neu
		helpi = (Count / (350 * 5)) + 1;						// Neu
		helpi = helpi * (350 * 5 );						// Neu
		daten = malloc(sizeof(long) * helpi);					// Neu
 		//daten = malloc(sizeof(long) * Count);
 	}
 	//else														//Neu
 		//daten = Data;

//udp_retry_get_data:

 	msize = sizeof(long);
 	pointer = 0;
 	start = StartIndex;

	h1.timeout = D->timeout;		//Neu
	h1.link= D->hostlink;				//Neu
	strcpy(h1.password, D->passwd);	//Neu

 	h1.befehl   = k_GetData;
 	h1.value[0] = dsp_typ;
 	h1.value[1] = DataNo | (5 << 16);
 	h1.value[2] = start;
 	h1.value[3] = Count;
 	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) return;//(1);//255 bei M.G.
 	if(getarray (daten, Count, msize, D->sockfd, (struct sockaddr *) &(D->server), 5, &my_time))
	{
		free(daten);   //Neu
		return;// 255;    //Neu
	}

 	D->fehlercode = 0;

 	if (dsp_typ != typ)
 	{
 		switch (typ)
 		{
			//TODO: case 1, 6, 20
 			case 2:				// 32bit int
 				plong = Data;
 				if (dsp_typ == 5)
 				{
 					pfloat =  (float *)  daten;
 					for (l1=0; l1<Count  ; l1++)
 					{
 						plong[l1] = (long)(pfloat [l1]);
					}
 				}
 				break;
 			case 5:				// 32bit float
 				pfloat = Data;
 				if (dsp_typ == 2)
 				{
 					plong =  (long *)  daten;
 					for (l1=0; l1<Count  ; l1++)
 					{
 						pfloat[l1] = (float)(plong[l1]);
					}
 				}
 				break;
			default:;

 		}
 		free(daten);
 	}
	else
	{
		memcpy(Data,daten,Count * sizeof(long));   //Neu
		free(daten);							//Neu
	}
	return ;//(0) ;
}

short Get_Data_fastshort(long DataNo,void *Data, short typ,  long StartIndex, long Count)
{
	int l1;
	short dsp_typ = 0;
	char *daten;
	UDPH1 h1;
	UDPR4 r4;
	struct timeval my_time;

	long  *plong;
	float *pfloat;

	int res,versuche,blockcount;

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	versuche = 0;

	strcpy(h1.password, D->passwd);
	h1.link = D->hostlink;

	blockcount = ((Count - 1) / 350) + 1 ;

	daten = malloc(sizeof(long) * blockcount * 350);

	if (daten == NULL)
	{
		return (k_Data_is_to_small);
	}

	h1.befehl   = k_GetDatasmall;
	h1.timeout = D->timeout;
	h1.value[0] = dsp_typ;
	h1.value[1] = DataNo | (blockcount << 16);
	h1.value[2] = StartIndex;
	h1.value[3] = Count;
	h1.link = 1;

//Get_Data_fastshort_nochmal:;
	res = do_udp_command_nx4(D->sockfd, &h1, &r4, &(D->server), &my_time, blockcount, daten, &dsp_typ);

	if (res < blockcount)
	{
		// hier stimmt was nicht
		// Dumme Fehlerlsung besser wre es nun nur noch den Rest amzuhollen.

/*		if ((fehlercode == k_Data_is_to_small) || (fehlercode == k_error_tcp_password)|| (fehlercode == k_timeout_to_link)|| (fehlercode == k_timeout_from_link))
		{
			return(255);
		}*/
/*		if (versuche < 5)							// da schon in do_udp_kommand_nx4
		{
			versuche = versuche + 1;
			goto Get_Data_fastshort_nochmal;		// so net!!!
		}*/
		free(daten);
		return((short)(D->fehlercode));		// wird vom do_udp_command_nx4 bereits richtig gesetzt
	}

	if (dsp_typ == 0)
	{
		free(daten);
		return (k_Data_is_to_small);
	}


	if ((dsp_typ == 3) || (dsp_typ == 4))
		dsp_typ = 2;

	switch (typ)
	{
		case 2:
			plong = Data;
			if (dsp_typ == 2)
				memcpy(plong,daten,Count * sizeof(long));
			if (dsp_typ == 5)
			{
				pfloat =  (float *)  daten;
				for (l1=0; l1<Count  ; l1++)
					plong[l1] = (long)(pfloat [l1]);
			}
		break;
		case 5:
			pfloat = Data;
			if (dsp_typ == 2)
			{
				plong =  (long *)  daten;
				for (l1=0; l1<Count  ; l1++)
					pfloat[l1] = (float)(plong[l1]);
			}
			if (dsp_typ == 5)
				memcpy(pfloat,daten, Count * sizeof(long));
		break;
	}
	free(daten);
	return(0);
}


void NAME (Set_Data) (long DataNo,void *Data, short typ,  long StartIndex, long Count)
{
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;  //Neu

	long l1 ;
	short res;
	short *pshort;
	long  *plong;
	float *pfloat;
	int pointer,start,mysize,retry;
	short dsp_typ;
	char *daten;
	long paket_index_dieses_Befehls;			// Neu 06.04.04

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	if (Count < 1) return;//(1); //0 bei M.G.

	D->fehlercode = 0;

	dsp_typ = (short)NAME (GetDataType) ((short)DataNo);

	if (dsp_typ == 0)
 	{
 		if (D->fehlercode == 0) D->fehlercode = k_Data_is_to_small;
 		if (D->fehlercode == k_DataZuKlein) D->fehlercode = k_Data_is_to_small;
 		return;//(255);
 	}

	if ((dsp_typ == 3) || (dsp_typ == 4))
 		dsp_typ = 2;

 	if (dsp_typ != typ)
 	{
 		daten = malloc(sizeof(long) * Count);
 		switch (typ)
 		{
 			case 1:
 				if (dsp_typ == 2)
 				{
 					pshort = (short *) Data;
 					plong =  (long *)  daten;
 					for (l1=0; l1<Count  ; l1++)
 					{
 						plong[l1] = pshort[l1];
 					}
 				}
 				if (dsp_typ == 5)
 				{
 					pshort = (short *) Data;
 					pfloat =  (float *)  daten;
 					for (l1=0; l1<Count  ; l1++)
 					{
 						pfloat [l1] = pshort[l1];
 					}
 				}
 			break;
 			case 2:
  			if (dsp_typ == 5)
  			{
  				plong = (long *) Data;
  				pfloat =  (float *)  daten;
  				for (l1=0; l1<Count  ; l1++)
  				{
  					pfloat [l1] = (float)(plong[l1]);
  				}
  			}
  			break;
 			case 5:
 				if (dsp_typ == 2)
 				{
 					pfloat =  (float *)  Data;
 					plong =  (long *)  daten;
 					for (l1=0; l1<Count  ; l1++)
 					{
 						plong[l1] = (long)(pfloat[l1]);
 					}
 				}
 				break;
  		}
	}
	else
		daten = (char *)Data;

	retry = 0;
	pointer = 0;
	start = StartIndex;

	mysize = sizeof(long);

	strcpy(h1.password, D->passwd); //neu
	h1.link = D->hostlink;			 //neu
	h1.timeout = D->timeout;     //neu

 	h1.befehl   = k_SetData;
 	h1.value[0] = dsp_typ;
 	h1.value[1] = DataNo | (5 << 16);
 	h1.value[2] = start;
 	h1.value[3] = Count;

 	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time))
 	{
 		if (D->fehlercode == k_DataZuKlein) D->fehlercode = k_Data_is_to_small;
 		if (dsp_typ != typ) free(daten);
 		return;//(1);
 	}

 	if (r1.status != k_ok)
 	{
 		if (D->fehlercode == k_DataZuKlein) D->fehlercode = k_Data_is_to_small;
 		if (dsp_typ != typ) free(daten);
 		return;//(1);
 	}

	paket_index_dieses_Befehls = D->udp_paket_count;			// Neu 06.04.04
 	if (setarray(daten, Count, sizeof(long),D->sockfd, (struct sockaddr *) &(D->server) , 5, &my_time))
 	{
 		if (dsp_typ != typ) free(daten);
 		return;//(1);
 	}
 	if (dsp_typ != typ) free(daten);

udp_retry:;

 	res = recvfromServer(D->sockfd, &r1, &my_time);

	// Hier mu berprft werden ob Server timeout sendet!!
			if (k_ok != res)
			{
				// Fehlerbehandlung

				if (res > 0)
				{
					D->fehlercode = res;
					return;//(255);
				}

				D->fehlercode = k_error_tcp_error;

				if (res == k_UDPtimeout)
				{
					D->fehlercode = k_error_tcp_timout_read;
				}
				if (retry == 0)
				{
					retry = 1;
					goto udp_retry;
				}
				// Falls alles schiefgegangen ist !!		// Neu 06.04.04
				if (D->ip_version >= 264 ||
					((D->ip_version >= 253) && (D->ip_version < 260)))
				{
					my_time.tv_sec = D->timeout / 1000;
					my_time.tv_usec = (D->timeout % 1000) * 1000;
					h1.befehl   = k_SetDataLastStatus;
					h1.timeout = D->timeout;
					h1.value[0] = paket_index_dieses_Befehls ;
					h1.value[1] = 0;
					h1.value[2] = 0;
				
					strcpy(h1.password , D->passwd);
					h1.link = D->hostlink;
					if (do_udp_command (D->sockfd, &h1, &r1, &(D->server), &my_time))
					{
						D->fehlercode = k_error_tcp_error;
						return;
					}
					if (r1.value[0] == 1)
					{
						// war ok
						D->fehlercode = 0;
						return;
					}					
					else
					{
						D->fehlercode = k_error_tcp_error;
					}
				}									// Neu 06.04.04 Ende				
				return;//(255);
			}
			D->fehlercode = 0;

			if (k_ok == res)
			{
				if (r1.status != k_ok)
				{
					//fehlercode = r1.status; this switch is taken out of the 'adwin32.c' it's just to be similar to windows
						switch (r1.status)
						{
							case 1:
								D->fehlercode = k_timeout_to_link;
							break;
							case 2:
								D->fehlercode = k_timeout_from_link;
							break;
							case 5:
								D->fehlercode = k_timeout_to_link;
							break;
							case -1:
							case -10:
							case -15:
							case -20:
							case -25:
							case -26:
							case -30:
							case -33:
							case -40:
							case -41:
								D->fehlercode = k_error_tcp_error;
							break;
							case -5:
								D->fehlercode = k_error_tcp_password;
							break;

							case -31:
								D->fehlercode = k_Data_is_to_small;
							break;

							case -32:
								D->fehlercode = k_Fifo_to_less_data;
							break;
						}
					return ;//255;
				}
			}
	return ;//(res) ;
}


//short Clear_Data(short Datanr)//--->soll auskommentiert werden, weil es das in der AktiveX nicht gibt
//{
//	UDPH1 h1;			//Neu
//	UDPR1 r1;			//Neu
//	struct timeval my_time;  //Neu
//
//	my_time.tv_sec  =  UDPTimeout / 1000;		//Neu
//	my_time.tv_usec = (UDPTimeout % 1000)*1000;	//Neu
//
//	h1.timeout = UDPTimeout;		//Neu
//	h1.link= hostlink;				//Neu
//	strcpy(h1.password, passwd);	//Neu
//
// 	h1.befehl   = k_3plus1;
// 	h1.value[0] = k_Clear_Data;
// 	h1.value[1] = Datanr;
// 	h1.value[2] = 0;
//
// 	if (do_udp_command(sockfd, &h1, &r1, &server, &my_time)) return(1);		//false bei M.G.
// 	return(0); //True bei M.G.
//}


//short CreateData(short datanr, short typ, long groesse, short ort, short art)//--->s. Clear_Data()
//{
//	UDPH1 h1;			//Neu
//	UDPR1 r1;			//Neu
//	struct timeval my_time;  //Neu
//
//	my_time.tv_sec  =  UDPTimeout / 1000;		//Neu
//	my_time.tv_usec = (UDPTimeout % 1000)*1000;	//Neu
//
//	h1.timeout = UDPTimeout;		//Neu
//	h1.link= hostlink;				//Neu
//	strcpy(h1.password, passwd);	//Neu
//
// 	h1.befehl   = k_CreateData;
// 	h1.value[0] = datanr;
// 	h1.value[1] = typ;
// 	h1.value[2] = groesse;
// 	h1.value[3] = art | (ort << 16);
//
// 	if (do_udp_command(sockfd, &h1, &r1, &server, &my_time )) return(1);		//255 bei M.G.
// 	return(0);
//}


void NAME (Data2File) (char *Filename, long DataNo, long Startindex, long Count, long Mode)
{
	FILE *file;
	void *data;

	file = fopen (Filename, Mode ? "ab" : "wb");
	if (file == NULL)
	{
		D->fehlercode = 200;
		return;
	}
	data = malloc (Count * sizeof (long));
	NAME (Get_Data) (DataNo, data, (short)(NAME (GetDataType)) ((short)DataNo), Startindex, Count);
	fwrite (data, 1, Count, file);
	free (data);
	fclose (file);
}

/******Boot*und*Processes****************************************************************************/

long get_filesize(FILE *fd)
{
	long size;

	fseek(fd, 0, SEEK_END);
	size = ftell(fd);
	rewind(fd);

	return (size);
}


void NAME (Boot) (char *dateiname)
//long int_NEUISERVER(char *dateiname, short Linkaddresse, long iboardsize)
{
	/* Returnwert <= 1000 => Fehler */

	int res,retry,filesize;
	FILE *fd;
	long pos_in_name,lauf;

	UDPH1 h1;			//Neu
	UDPR1 r1;			//Neu
	struct timeval my_time;  //Neu

	short help;

	lauf = strlen(dateiname);

	if ((dateiname[lauf-4] != '.') || ((dateiname[lauf-3] != 'B') && (dateiname[lauf-3] != 'b'))|| ((dateiname[lauf-2] != 'T') && (dateiname[lauf-2] != 't'))|| ((dateiname[lauf-1] != 'L') && (dateiname[lauf-1] != 'l')))
	{
		//keine MessageBox!!
		D->fehlercode = k_File_not_btl; //MS changed from k_File_not_binaer; //
		return;// 0;
	}


	//long return_wert;

	//return_wert = iboardsize;
	help = 0;

//    if (D->hostlink <= 0) return;//(1);

	pos_in_name = 0;
	while (dateiname[pos_in_name + 1] != '.') pos_in_name =  pos_in_name + 1;
	//Test if the filename is a valid name.
	//It tests only if there is a '9' or a '10' bevor the point(.)!!!
	if (dateiname[pos_in_name] == '9'|| ((dateiname[pos_in_name - 1] == '1') && (dateiname[pos_in_name] == '0')))
	{
		unsigned char *daten;
		retry = 0;

		my_time.tv_sec  =  D->timeout / 1000;		//Neu
		my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

		h1.timeout = D->timeout;		//Neu
		h1.link= D->hostlink;				//Neu
		strcpy(h1.password, D->passwd);	//Neu

		h1.befehl   = k_Boot;
		h1.value[0] = 1000; // Iboardsize

		if ((fd = fopen (dateiname, "rb")) == NULL)
		{
			//keine MessageBox!!
			D->fehlercode = k_File_not_found;//k_wrong_version;

			return;//(1); //0 bei M.G.
		}

 		filesize= get_filesize(fd);
 		h1.value[1] = 5 << 16;		// PaketCount
 		h1.value[2] = k_proc_t9;
 		h1.value[3] = filesize;

 		daten = (char *) malloc(filesize);
 		fread(daten, sizeof(char), filesize, fd);
 		fclose(fd);

 		if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time))
		{
			free(daten);
			return;//(1);
		}

 		if (setarray(daten, filesize, sizeof(char), D->sockfd, (struct sockaddr *) &(D->server), 5, &my_time))
 		{
 			free(daten);
 			return;//(1);  //255 bei M.G.
 		}
 		free(daten);

udp_retry:;
 		res = recvfromServer(D->sockfd, &r1, &my_time);

 		if (k_ok != res)
 		{
 			// Fehlerbehandlung

 			if (res > 0)
 			{
 				D->fehlercode = res;
 				return;//(1);   //0 bei M.G.
 			}

 			D->fehlercode = k_error_tcp_error;

 			if (res == k_UDPtimeout)
 			{
 				D->fehlercode = k_error_tcp_timout_read;
 			}
 			if (retry < 5)
 			{
 				retry = retry + 1;
 				goto udp_retry;
 			}
 			return;//(1);//0 bei M.G.
 		}

 		if (r1.status != k_ok)
 		{
			//fehlercode = r1.status; this switch is taken out of the 'adwin32.c' it's just to be similar to windows
				switch (r1.status)
				{
					case 1:
						D->fehlercode = k_timeout_to_link;
					break;
					case 2:
						D->fehlercode = k_timeout_from_link;
					break;
					case 5:
						D->fehlercode = k_timeout_to_link;
					break;
					case -1:
					case -10:
					case -15:
					case -20:
					case -25:
					case -26:
					case -30:
					case -33:
					case -40:
					case -41:
						D->fehlercode = k_error_tcp_error;
					break;
					case -5:
						D->fehlercode = k_error_tcp_password;
					break;

					case -31:
						D->fehlercode = k_Data_is_to_small;
					break;

					case -32:
						D->fehlercode = k_Fifo_to_less_data;
					break;
				}
 			return;//(1); //0 bei M.G.
 		}
	}
	else
		D->fehlercode = k_File_not_valid;
	return;//(0); //8000 bei M.G.!!?
}

//24.11.2003 K.Sp. auskommentiert
//long ADboot(char *dateiname, short Linkaddresse, long iboardsize)
//{
//	return int_NEUISERVER(dateiname, Linkaddresse, iboardsize);
//}


void NAME (Load_Process) (char *dateiname)
//short int_Starte(char  *dateiname, short Linkaddresse)
{
	/* Erfolgreich geladen 0
     nicht erfolgreich   1 */
 	int res,filesize,retry;
	int lauf;
 	FILE *fd;
 	char *daten;

	UDPH1 h1;			//Neu
	UDPR1 r1;			//Neu
	struct timeval my_time;  //Neu


	lauf = strlen(dateiname);

	if ((dateiname[lauf-4] != '.') || ((dateiname[lauf-3] != 'T') && (dateiname[lauf-3] != 't')))
	{
		D->fehlercode = k_File_not_binaer;
		return;// 0;
	}

 	retry = 0;

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	h1.timeout = D->timeout;		//Neu
	h1.link= D->hostlink;				//Neu
	strcpy(h1.password, D->passwd);	//Neu

 	h1.befehl   = k_LadeBin;

 	if ((fd = fopen (dateiname, "rb")) == NULL)
 	{
 		D->fehlercode = k_File_not_found;
 		return;//(1); //0 bei M.G.
 	}

 	filesize= get_filesize(fd);
 	daten = (char *) malloc(filesize);
 	fread(daten, sizeof(char), filesize, fd);
 	fclose(fd);

 	h1.value[1] = 5 << 16;		// PaketCount
 	h1.value[2] = k_proc_t9;
 	h1.value[3] = filesize;

 	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time))
	{
		free(daten);
		return;//(1); //0 bei M.G.
	}

 	if (setarray(daten, filesize, sizeof(char), D->sockfd,(struct sockaddr *) &(D->server), 5, &my_time))
 	{
 		free(daten);
 		return;//(1); //255 bei M.G.
 	}
 	free(daten);

udp_retry:;

 	res = recvfromServer(D->sockfd, &r1, &my_time);
 	if (k_ok != res)
 	{
 		// Fehlerbehandlung

 		if (res > 0)
 		{
 			D->fehlercode = res;
 			return;//(1); //0 bei M.G.
 		}

 		D->fehlercode = k_error_tcp_error;

 		if (res == k_UDPtimeout)
 		{
 			D->fehlercode = k_error_tcp_timout_read;
 		}
 		if (retry < 5)
 		{
 			retry = retry + 1;
 			goto udp_retry;
 		}
 		return;//(1); //0 bei M.G.
 	}
 	if (r1.status != k_ok)
 	{
		if (r1.status == -30)	D->fehlercode = k_wrong_version;
		else
		{
			//fehlercode = r1.status; this switch is taken out of the 'adwin32.c' it's just to be similar to windows
				switch (r1.status)
				{
					case 1:
						D->fehlercode = k_timeout_to_link;
					break;
					case 2:
						D->fehlercode = k_timeout_from_link;
					break;
					case 5:
						D->fehlercode = k_timeout_to_link;
					break;
					case -1:
					case -10:
					case -15:
					case -20:
					case -25:
					case -26:
					case -30:
					case -33:
					case -40:
					case -41:
						D->fehlercode = k_error_tcp_error;
					break;
					case -5:
						D->fehlercode = k_error_tcp_password;
					break;

					case -31:
						D->fehlercode = k_Data_is_to_small;
					break;

					case -32:
						D->fehlercode = k_Fifo_to_less_data;
					break;
				}
		}
 		return;//(1); //0 bei M.G.
 	}
 	D->fehlercode = 0;
 	return;//(0); //1 bei M.G.
}

//24.11.2003 K.Sp. auskommentiert
//short ADBload(char *dateiname, short Linkaddresse)
//{
//	return int_Starte(dateiname, Linkaddresse);
//}


void NAME (Start_Process) (long ProcessNo)
{
	UDPH1 h1;			//Neu
	UDPR1 r1;			//Neu
	struct timeval my_time;  //Neu

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	h1.timeout = D->timeout;		//Neu
	h1.link= D->hostlink;				//Neu
	strcpy(h1.password, D->passwd);	//Neu

 	h1.befehl   = k_3plus1;
 	h1.value[0] = k_Start;
 	h1.value[1] = ProcessNo; //klren--> anders zu DLL
 	h1.value[2] = 0;
 	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) return;//(1);//false
 	return;//(0);
}


void NAME (Stop_Process) (long ProcessNo)
{
	UDPH1 h1;			//Neu
	UDPR1 r1;			//Neu
	struct timeval my_time;  //Neu

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	h1.timeout = D->timeout;		//Neu
	h1.link= D->hostlink;				//Neu
	strcpy(h1.password, D->passwd);	//Neu

 	h1.befehl   = k_3plus1;
 	h1.value[0] = k_Stop;
 	h1.value[1] = ProcessNo;
 	h1.value[2] = 0;
 	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) return;//(1);//false
 	return;//(0);
}


void NAME (Clear_Process) (long ProcessNo)
{
	UDPH1 h1;			//Neu
	UDPR1 r1;			//Neu
	struct timeval my_time;  //Neu

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	h1.timeout = D->timeout;		//Neu
	h1.link= D->hostlink;				//Neu
	strcpy(h1.password, D->passwd);	//Neu

 	h1.befehl   = k_3plus1;
 	h1.value[0] = k_ClearProcess;
 	h1.value[1] = ProcessNo;
 	h1.value[2] = 0;

 	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) return;//(1);
 	return;//(0);
 }

long NAME (Process_Status) (long ProcessNo)
{
	return NAME (Get_Par) (-100 + ProcessNo);
}

long NAME (Get_GlobalDelay) (long ProcessNo)
{
	return NAME (Get_Par) (-90 + ProcessNo);
}

void NAME (Set_GlobalDelay) (long ProcessNo, long GlobalDelay)
{
	NAME (Set_Par) (-90 + ProcessNo, GlobalDelay);
}


long NAME (Workload) () //long Workload(long Priority)
{
	UDPH1 h1;			//Neu
	UDPR1 r1;			//Neu
	struct timeval my_time; //Neu
	unsigned short wert;
	long wert1,zeit1;
	float help;

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	h1.timeout = D->timeout;		//Neu
	h1.link= D->hostlink;				//Neu
	strcpy(h1.password, D->passwd);	//Neu


	h1.befehl = k_Auslastung;
	h1.value[0] = k_Auslastung;
	h1.value[1] = 0;
	h1.value[2] = 0;

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) return(100); //255 bei M.G.

	if (r1.value[2] == 0)
	{
		wert1 = r1.value[0];
		zeit1 = r1.value[1];
		if (zeit1 == 0) wert = 100;
		else
		{
			help = (float)wert1 / (float) zeit1;
			wert = (unsigned short)(100. * ((0.25 - help) / 0.25)) ;
		}
	}
	else{
		wert = (unsigned short)(r1.value[0]);
	}
	return(wert);
}

long NAME (Free_Mem) (long Mem_Spec)
{
	UDPH1 h1;			//Neu
	UDPR1 r1;			//Neu
	struct timeval my_time;  //Neu

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	h1.timeout = D->timeout;		//Neu
	h1.link= D->hostlink;				//Neu
	strcpy(h1.password, D->passwd);	//Neu

 	h1.befehl   = k_3plus1;
 	h1.value[0] = k_MemoryAll;
 	h1.value[1] = Mem_Spec;
 	h1.value[2] = 0;

 	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) 
		return(1); //255
 	switch (Mem_Spec)
 	{
	 	case 1:	return(r1.value[0]*6);	// PM
 		case 3:	return(r1.value[0]*4);	// DM
 		case 4:	return(r1.value[0]*4);	// DRAM
 		default: return(0);
 	}
}

/****Get_Last_Error*******************************************************************************/

long NAME (Get_Last_Error) ()
{
	long res;
	res = D->fehlercode;
	D->fehlercode = 0;
	return res;
}

long NAME (Get_Last_Error_Text) (long errorcode,char *text,long textlaenge)
{
	int count;
	char Buffer[200];

	switch(errorcode)
	{
		case 0:
			if (english)
				strcpy(Buffer,k_error_0_e);
			else
				strcpy(Buffer,k_error_0_d);
		break;
		/*case 1:
			if (english)
				strcpy(Buffer,k_error_1_e);
			else
				strcpy(Buffer,k_error_1_d);
		break;

		case 2:
			if (english)
				strcpy(Buffer,k_error_2_e);
			else
				strcpy(Buffer,k_error_2_d);
		break;

		case 3:
			if (english)
				strcpy(Buffer,k_error_3_e);
			else
				strcpy(Buffer,k_error_3_d);
		break;

		case 4:
			if (english)
				strcpy(Buffer,k_error_4_e);
			else
				strcpy(Buffer,k_error_4_d);
		break;

		case 10:
			if (english)
				strcpy(Buffer,k_error_10_e);
			else
				strcpy(Buffer,k_error_10_d);
		break;

		case 11:
			if (english)
				strcpy(Buffer,k_error_11_e);
			else
				strcpy(Buffer,k_error_11_d);
		break;

		case 15:
			if (english)
				strcpy(Buffer,k_error_15_e);
			else
				strcpy(Buffer,k_error_15_d);
		break;*/

		case 20:
			if (english)
				strcpy(Buffer,k_error_20_e);
			else
				strcpy(Buffer,k_error_20_d);
		break;

		case 100:
			if (english)
				strcpy(Buffer,k_error_100_e);
			else
				strcpy(Buffer,k_error_100_d);
		break;

		/*case 101:
			if (english)
				strcpy(Buffer,k_error_101_e);
			else
				strcpy(Buffer,k_error_101_d);
		break;*/

		case 102:
			if (english)
				strcpy(Buffer,k_error_102_e);
			else
				strcpy(Buffer,k_error_102_d);
		break;

		case 200:
			if (english)
				strcpy(Buffer,k_error_200_e);
			else
				strcpy(Buffer,k_error_200_d);
		break;

		/*case 201:
			if (english)
				strcpy(Buffer,k_error_201_e);
			else
				strcpy(Buffer,k_error_201_d);
		break;*/

		case 202:
			if (english)
				strcpy(Buffer,k_error_202_e);
			else
				strcpy(Buffer,k_error_202_d);
		break;

		case 203:
			if (english)
				strcpy(Buffer,k_error_203_e);
			else
				strcpy(Buffer,k_error_203_d);
		break;

		/*case 204:
			if (english)
				strcpy(Buffer,k_error_204_e);
			else
				strcpy(Buffer,k_error_204_d);
		break;

		case 1000:
			if (english)
				strcpy(Buffer,k_error_1000_e);
			else
				strcpy(Buffer,k_error_1000_d);
		break;

		case 1001:
			if (english)
				strcpy(Buffer,k_error_1001_e);
			else
				strcpy(Buffer,k_error_1001_d);
		break;

		case 1002:
			if (english)
				strcpy(Buffer,k_error_1002_e);
			else
				strcpy(Buffer,k_error_1002_d);
		break;*/

		/*case 1003:
			if (english)
				strcpy(Buffer,k_error_1003_e);
			else
				strcpy(Buffer,k_error_1003_d);
		break;*/

		case 2000:
			if (english)
				strcpy(Buffer,k_error_2000_e);
			else
				strcpy(Buffer,k_error_2000_d);
		break;

		case 2001:
			if (english)
				strcpy(Buffer,k_error_2001_e);
			else
				strcpy(Buffer,k_error_2001_d);
		break;

		case 2002:
			if (english)
				strcpy(Buffer,k_error_2002_e);
			else
				strcpy(Buffer,k_error_2002_d);
		break;

		/*case 3000:
			if (english)
				strcpy(Buffer,k_error_3000_e);
			else
				strcpy(Buffer,k_error_3000_d);
		break;

		case 3001:
			if (english)
				strcpy(Buffer,k_error_3001_e);
			else
				strcpy(Buffer,k_error_3001_d);
		break;*/

		default:
			if (english)
				strcpy(Buffer,k_error_un_e);
			else
				strcpy(Buffer,k_error_un_d);
		break;

	}
	count = strlen(Buffer) + 1;
	if (count > textlaenge) count = textlaenge;

	textlaenge = count;

	memcpy(text,Buffer,sizeof(char) * (textlaenge));

	text[textlaenge - 1] = 0;

	return(textlaenge - 1);
}
/**********Language******************************************************************************/

void Set_Language(long language)
{
	switch (language)
	{
	/*case 0:
		{
			long langua;
			langua = GetSystemDefaultLangID();
			if ((langua == 0x407) || (langua == 0x807) || (langua == 0xC07))
			{
				english = false;
			}
			else
			{
				english = true;
			}
		}
		break;*/
	case 1:
		english = 1;
		break;
	case 2:
		english = 0;
		break;
	default:
		english = 1;
		break;
	}
	return;
}

/************Version, Prozessor**************************************************************/

long intern_Get_Version_TcpIp()
{

	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	my_time.tv_sec  =  D->timeout / 1000;
	my_time.tv_usec = (D->timeout % 1000)*1000000;

	h1.befehl   = k_armVersion;
	h1.timeout = D->timeout;
	h1.value[0] = k_armVersion;
	h1.value[1] = 0;
	h1.value[2] = 0;
	h1.value[3] = 0;

	strcpy(h1.password, D->passwd);
	h1.link= D->hostlink;

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time)) return 0;

	D->fehlercode = 0;

	return(r1.value[0]);
}

long NAME (Test_Version) ()
{
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	my_time.tv_sec = D->timeout / 1000;
	my_time.tv_usec = (D->timeout % 1000) * 1000;

	h1.befehl   = k_TestVersion;
	h1.timeout = D->timeout;
	strcpy(h1.password , D->passwd);
	h1.link = D->hostlink;

	if (do_udp_command(D->sockfd, &h1,&r1,&(D->server),&my_time )) return 3;
//	Prozessor= r1.value[1];
//	Kartentyp= r1.value[2];
//	Treiberversion = r1.value[3];

#ifndef _MYUDPTEST
	D->fehlercode = 0;
#endif
	return 0;
}

long NAME (Processor_Type) ()
{
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	my_time.tv_sec = D->timeout / 1000;
	my_time.tv_usec = (D->timeout % 1000) * 1000;

	h1.befehl   = k_TestVersion;
	h1.timeout = D->timeout;
	strcpy(h1.password , D->passwd);
	h1.link = D->hostlink;

	if (do_udp_command(D->sockfd, &h1,&r1, &(D->server), &my_time )) return 0;

#ifndef _MYUDPTEST
	D->fehlercode = 0;
#endif
	return r1.value[0];
}


void NAME (GetData_Long) (long DataNo, long *Data, long Startindex, long Count)
{
	NAME (Get_Data) (DataNo, Data, 2, Startindex, Count);
}

void NAME (GetData_Float) (long DataNo, float *Data, long Startindex, long Count)
{
	NAME (Get_Data) (DataNo, Data, 5, Startindex, Count);
}

void NAME (SetData_Long) (long DataNo, long *Data, long Startindex, long Count)
{
	NAME (Set_Data) (DataNo, Data, 2, Startindex, Count);
}

void NAME (SetData_Float) (long DataNo, float *Data, long Startindex, long Count)
{
	NAME (Set_Data) (DataNo, Data, 5, Startindex, Count);
}

void NAME (SetFifo_Long) (long FifoNo, long *Data, long Count)
{
	NAME (Set_Fifo) (FifoNo, 2, Data, Count);
}
void NAME (SetFifo_Float) (long FifoNo, float *Data, long Count)
{
	NAME (Set_Fifo) (FifoNo, 5, Data, Count);
}
void NAME (GetFifo_Long) (long FifoNo, long *Data, long Count)
{
	NAME (Get_Fifo) (FifoNo, 2, Data, Count);
}
void NAME (GetFifo_Float) (long FifoNo, float *Data, long Count)
{
	NAME (Get_Fifo) (FifoNo, 5, Data, Count);
}

long NAME (String_Length) (long DataNo)
{
	long value;

	NAME (Get_Data) (DataNo, &value, 2, 1, 1);
	if (D->fehlercode != 0)
		return -1;

	return value;
}

long NAME (GetData_String) (long DataNo, char *Data, long MaxCount)
{
	long *values;
	long lauf, res;

	values = (long *) malloc ((MaxCount + 2) * sizeof(long));

	NAME (Get_Data) (DataNo, values, 2, 1, MaxCount + 1);
	if (D->fehlercode != 0)
	{
		// ging nicht
		free (values);
		return -1;
	}
	res = values[0];
	if (MaxCount - 1 < res) res = MaxCount - 1;

	for (lauf = 0; lauf < res; lauf ++)
	{
		Data[lauf] = (char) values[lauf + 1];
	}

	Data[lauf] = 0;

	free (values);
	return res;
}

void NAME (SetData_String) (long DataNo, char *Data)
{
	unsigned long *values;
	long lauf,anzahl;


	anzahl = strlen (Data);
	values = (long*)malloc ((anzahl + 2) * sizeof (long));

	values[0] = anzahl;

	for (lauf = 0; lauf < anzahl; lauf ++)
		values[lauf + 1] = (unsigned char) Data[lauf];

	values[lauf + 1] = 0;

	NAME (Set_Data) (DataNo, values, 2, 1, anzahl + 2);

	free (values);
}

long NAME (Data_Length) (long DataNo)
{
	int lh;
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	h1.befehl   = k_3plus1;
	h1.timeout = D->timeout;
	h1.value[0] = k_GetDataLength;
	h1.value[1] = DataNo;
	strcpy (h1.password, D->passwd);
	h1.link = D->hostlink;

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time ))
	{
		return 0;														// gendert 08.05.02
	}
	lh = r1.value[0];
	if (lh > 0) lh = lh - 1;

	return lh;
}

int NAME (Clear_Data) (long DataNo)
{
	UDPH1 h1;
	UDPR1 r1;
	struct timeval my_time;

	my_time.tv_sec  =  D->timeout / 1000;		//Neu
	my_time.tv_usec = (D->timeout % 1000)*1000000;	//Neu

	h1.befehl   = k_3plus1;
	h1.timeout = D->timeout;
	h1.value[0] = k_Clear_Data;
	h1.value[1] = DataNo;
	h1.value[2] = 0;

	strcpy(h1.password, D->passwd);
	h1.link = D->hostlink;

	if (do_udp_command(D->sockfd, &h1, &r1, &(D->server), &my_time ))
	{
		return 0;
	}

	return 1;
}
