/********************************************************************
 * Simple Linux test suite for ADwin systems 
 *
 * written by Thomas Boehne (TBoehne@ADwin.de)
 * Jaeger Messtechnik GmbH
 * http://www.ADwin.de/
 * $Id: test.c,v 1.3 2004/07/29 07:48:16 cvs Exp $
 */ 

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libadwin.h>
#include <libadwin/bits.h>

// do {} while(0); construct makes PRINTERROR a single statment (important) TB
#define PRINTERROR \
  do { \
    int errNumber = Get_Last_Error(); \
    printf("%s (Error number %d)\n", Get_Last_Error_Text(errNumber), errNumber); \
  } while (0); 

char *processorTypeToString(int pType) {
  switch (pType) {
  case k_proc_t2:  return "T2";
  case k_proc_t4:  return "T4";
  case k_proc_t5:  return "T5";
  case k_proc_t8:  return "T8";
  case k_proc_t9:  return "T9";
  case k_proc_t10: return "T10";
  //  case k_proc_t11: return "T11"; // not supported yet
  }
  return "unknown";
}

int stringToProcessorType(char *string) {
  if (! (strncmp(string, "T2", 2) && strncmp(string, "t2", 2)))
    return k_proc_t2;
  if (! (strncmp(string, "T4", 2) && strncmp(string, "t4", 2)))
    return k_proc_t4;
  if (! (strncmp(string, "T5", 2) && strncmp(string, "t5", 2)))
    return k_proc_t5;
  if (! (strncmp(string, "T8", 2) && strncmp(string, "t8", 2)))
    return k_proc_t8;
  if (! (strncmp(string, "T9", 2) && strncmp(string, "t9", 2)))
    return k_proc_t9;
  if (! (strncmp(string, "T10", 3) && strncmp(string, "t10", 3)))
    return k_proc_t10;
  // T11 not supported yet
  //  if (! (strncmp(string, "T11", 3) && strncmp(string, "t11", 3)))
  //    return k_proc_t11;
  return 0;
}

char* setAdwindir(char *adwindir) {
  FILE *configFile;

  configFile = fopen("/etc/adwin/ADWINDIR", "r");
  if (configFile == NULL) {
    printf("Unable to open /etc/adwin/ADWINDIR\n");
    exit(1);
  }

  if (fscanf(configFile, "%254s", adwindir)  == EOF) {
    printf("/etc/adwin/ADWINDIR is empty\n");
    exit(1);
  }

  fclose(configFile);
  return adwindir;
}

char* setBtlPath(int processorType, char *adwindir, char *btlPath) {
  sprintf(btlPath, "%s/share/btl/adwin%d.btl", adwindir, processorType);
  return btlPath;
}

char* setProcessPath(int processorType, char *adwindir, char *processPath) {
  sprintf(processPath, "%s/share/examples/simple_start/inccount.t%x1", adwindir, processorType);
  return processPath;
}

int main(int argc, char *argv[]) {
  char answer = 'h'; // show help screen initially
  int deviceNumber = 1;
  int processorType = k_proc_t10;
  char adwindir[255], btlPath[285], processPath[285], tmpString[10];
  
  printf("ADWINDIR is set to: %s\n", setAdwindir(adwindir));
  setBtlPath(processorType, adwindir, btlPath);
  setProcessPath(processorType, adwindir, processPath);
  
  while (answer != 'x' && answer != 'X' && answer != 10) {
    switch (answer) {
    case 'H': case 'h':
      printf("H = print this help screen\n");
      printf("C = change device number (current: 0x%x)\n", deviceNumber);
      printf("T = change processor type (current: %s)\n", 
	     processorTypeToString(processorType));
      printf("B = boot ADwin system\n");
      printf("L = load process to ADwin system\n");
      printf("S = start process on ADwin system\n");
      printf("D = display counter from ADwin system\n");
      printf("R = reset counter on ADwin system\n");
      printf("X = exit program\n");
      break;
      
    case 'C': case 'c': 
      printf("Please enter Devicenumber: ");
      scanf("%s", tmpString);
      deviceNumber = strtol(tmpString, NULL, 0);
	  // Set_DeviceNo is called in case 'B' 
	  // (otherwise problem if default DeviceNo is used)
      break;
      
    case 'T': case 't': 
      do {
	// printf("Please enter processor type (T2, T4, T5, T8, T9, T10 or T11): ");
	printf("Please enter processor type (T2, T4, T5, T8, T9 or T10): ");
	scanf("%s", tmpString);
	processorType = stringToProcessorType(tmpString);
      } while (processorType == 0);
      printf("Using BTL %s\n", setBtlPath(processorType, adwindir, btlPath));
      printf("Using process %s\n", setProcessPath(processorType, adwindir, processPath));
      break;
      
    case 'B': case 'b': 
      Set_DeviceNo(deviceNumber);
      Boot(btlPath);
      PRINTERROR;
      break;
      
    case 'L': case 'l': 
      Load_Process(processPath);
      PRINTERROR;
      break;
      
    case 'S': case 's':
      Start_Process(1);
      PRINTERROR;
      break;
      
    case 'D': case 'd':
      printf("Value is %ld\n", Get_Par(1));
      PRINTERROR;
      break;
      
    case 'R': case 'r':
      Set_Par(1, 0);
      printf("Counter reseted\n");
      PRINTERROR;
      break;
    }
    printf("Your choice: ");
    do scanf("%c", &answer); while (answer == 10);
  }
  return(0);
}
